# Copyright (C) 2020-2021 Hatching B.V
# All rights reserved.

from io import StringIO, BytesIO
from triage.pagination import Paginator
from requests import Request, Session, exceptions, utils

import binascii
import json
import os
import sys

version = 'alpha'

class Client:
    def __init__(self, token, root_url='https://api.tria.ge'):
        self.token = token
        self.root_url = root_url.rstrip('/')

    def _new_request(self, method, path, j=None, b=None, headers={}):
        headers = {
            'Authorization': 'Bearer {0}'.format(self.token),
            'User-Agent': 'Triage Python Client/{0} Python/{1}'
                .format(version, sys.version.split(' ')[0]),
            **headers
        }
        if j:
            return Request(method, self.root_url + path, data=json.dumps(j), headers=headers)
        return Request(method, self.root_url + path, data=b, headers=headers)

    def _req_file(self, method, path):
        r = self._new_request(method, path)
        with Session() as s:
            return s.send(r.prepare()).content

    def _req_json(self, method, path, data=None):
        if data is None:
            r = self._new_request(method, path, data)
        else:
            r = self._new_request(method, path, data,
                headers={'Content-Type': 'application/json'})

        try:
            with Session() as s:
                res = s.send(r.prepare())
                res.raise_for_status()
                return res.json()
        except exceptions.HTTPError as err:
            raise ServerError(err)

    def submit_sample_file(self, filename, file, interactive=False,
                           profiles=[], password=None):
        """
        Submit a file for analysis on Triage.

        Parameters:
            filename (str):
                The name of the file
            file (file):
                Object with a read method (e.g. open("file.exe", "r"))
            interactive (bool):
                Whether to select files for analysis; defaults to False
            profiles (list):
                Select a profile for every individiual target
                [
                    {
                        "pick": "file1.exe", // for archives
                        "profile": "w7_long"
                    }
                ]
            password (str):
                Password for the archive sample
        Returns:
            response (dict):
                {
                    'id': '200923-4zhlr84m42',
                    'status': 'pending',
                    'kind': 'file',
                    'filename': 'test.exe',
                    'private': False,
                    'submitted': '2020-09-23T07:26:26Z'
                }
        """
        d = {
            'kind': 'file',
            'interactive': interactive,
            'profiles': profiles,
        }
        if password:
            d['password'] = password
        body, content_type = encode_multipart_formdata({
            '_json': json.dumps(d),
            'file': (filename, file),
        })
        r = self._new_request('POST', '/v0/samples', b=body,
            headers={"Content-Type": content_type}
        )
        try:
            with Session() as s:
                res = s.send(r.prepare())
                res.raise_for_status()
                return res.json()
        except exceptions.HTTPError as err:
            raise ServerError(err)

    def submit_sample_url(self, url, interactive=False, profiles=[]):
        """
        Submit a url for analysis on Triage.

        Parameters:
            url (str): The url to analyse
            interactive (bool):
                Whether to select files for analysis; defaults to False
            profiles (list):
                Select a profile for every individual target
                [
                    {
                        "profile": "w7_long"
                    }
                ]
        Returns:
            response (dict):
                {
                    'id': '200923-esktrbc2a6',
                    'status': 'pending',
                    'kind': 'url',
                    'url': 'http://www.google.com',
                    'private': False,
                    'submitted': '2020-09-23T07:51:45Z'
                }
        """
        return self._req_json('POST', '/v0/samples', {
            'kind': 'url',
            'url': url,
            'interactive': interactive,
            'profiles': profiles,
        })

    def set_sample_profile(self, sample_id, profiles):
        """
        Set profile for a sample, if the sample has been submitted in
        interactive mode.

        Parameters:
            sample_id (str): The id of the sample
            profiles (list):
                Select a profile for every individiual target
                [
                    {
                        "pick": "file.exe", // for archives
                        "profile": "w7_long"
                    }
                ]
        Returns:
            response (dict):
                {}, empty dict
        """
        return self._req_json('POST', '/v0/samples/%s/profile' % sample_id, {
            'auto': False,
            'profiles': profiles,
        })

    def set_sample_profile_automatically(self, sample_id, pick=[]):
        """
        Set profile for a sample automatically, if the sample has been
        submitted in interactive mode.

        Parameters:
            sample_id (str): The id of the sample
            pick (list):
                [
                    "file1.exe",
                    "file2.exe"
                ]

        Returns:
            response (dict):
                {}, empty dict
        """
        return self._req_json('POST', '/v0/samples/%s/profile' % sample_id, {
            'auto': True,
            'pick': pick,
        })

    def owned_samples(self, max=20):
        """
        Returns a Paginator object with owned samples.

        Parameters:
            max (int): The maximum amount of samples to return

        Returns:
            Paginator (object):
                Loop over this object to get the samples
        """
        return Paginator(self, '/v0/samples?subset=owned', max)

    def public_samples(self, max=20):
        """
        Returns a Paginator object with public samples.

        Parameters:
            max (int): The maximum amount of samples to return

        Returns:
            Paginator (object):
                Loop over this object to get the samples
        """
        return Paginator(self, '/v0/samples?subset=public', max)

    def sample_by_id(self, sample_id):
        """
        Returns a sample object.

        Parameters:
            sample_id (str): The id of the sample

        Returns:
            Sample (object):
                {
                    'id': '200923-hb5gvebega',
                    'status': 'static_analysis',
                    'kind': 'url',
                    'url': 'http://google.com',
                    'private': False,
                    'tasks': [
                        {
                            'id': 'static1',
                            'status': 'reported'
                        }
                    ],
                    'submitted': '2020-09-23T08:04:58Z'
                }
        """
        return self._req_json('GET', '/v0/samples/{0}'.format(sample_id))

    def get_sample_file(self, sample_id):
        """
        Return the sample file.

        Parameters:
            sample_id (str): The id of the sample

        Returns:
            response (file):
                File object.
        """
        return self._req_file("GET", "/v0/samples/{0}/sample".format(sample_id))

    def delete_sample(self, sample_id):
        """
        Delete a sample.

        Parameters:
            sample_id (str): The id of the sample

        Returns:
            response (dict):
                {}, empty dict
        """
        return self._req_json('DELETE', '/v0/samples/{0}'.format(sample_id))

    def search(self, query, max=20):
        """
        Returns a Paginator object with search samples.

        Parameters:
            query (str): The search query
            max (int): The maximum amount of samples to return

        Returns:
            Paginator (object):
                Loop over this object to get the samples
        """
        params = utils.quote(query)
        return Paginator(self, '/v0/search?query={0}'.format(params), max)

    def static_report(self, sample_id):
        """
        Return a static report.

        Parameters:
            sample_id (str): The id of the sample

        Returns:
            response (dict):
                {
                    'version': '0.2',
                    'sample': {},
                    'task': {},
                    'analysis': {},
                    'files': None,
                    'unpack_count': 0,
                    'error_count': 0
                }
        """
        return self._req_json(
            'GET', '/v0/samples/{0}/reports/static'.format(sample_id)
        )

    def overview_report(self, sample_id):
        """
        Return an overivew of the sample.

        Parameters:
            sample_id (str): The id of the sample

        Returns:
            response (dict):
                {
                    'version': '0.2.2',
                    'sample': {},
                    'task': {},
                    'analysis': {},
                    'targets': {},
                    'extracted': {},
                    'errors': {} (omitted when empty)
                }
        """
        return self._req_json(
            'GET', '/v1/samples/{0}/overview.json'.format(sample_id)
        )

    def kernel_report(self, sample_id, task_id):
        """
        Return kernel output.

        Parameters:
            sample_id (str): The id of the sample
            task_id (str): The id of the task

        Returns:
            List of json entries
            [
                {"kind": ".."},
                {"kind": ".."},
                {"kind": ".."},
            ]
        """
        overview = self.overview_report(sample_id)
        for t in overview["tasks"]:
            if t["name"] == task_id:
                task = t
                break
        else:
            raise ValueError("Task does not exist")

        if "windows" in task["platform"]:
            r =  self._new_request(
                'GET', '/v0/samples/{0}/{1}/logs/onemon.json'.format(
                    sample_id, task_id)
            )
        elif "linux" in task["platform"]:
            r =  self._new_request(
                'GET', '/v0/samples/{0}/{1}/logs/stahp.json'.format(
                    sample_id, task_id)
            )
        else:
            raise ValueError("Platform not supported")

        with Session() as s:
            res = s.send(r.prepare())
            res.raise_for_status()
            for entry in res.content.split(b"\n"):
                if entry.strip() == b"":
                    break
                yield json.loads(entry)

    def task_report(self, sample_id, task_id):
        """
        Return a task report.

        Parameters:
            sample_id (str): The id of the sample
            task_id (str): The id of the task

        Returns:
            response (dict):
                {
                    'version': '0.2.2',
                    'sample': {},
                    'task': {},
                    'errors': [],
                    'analysis': {},
                    'signatures': [],
                    'network': {}
                }
        """
        return self._req_json(
            'GET', '/v0/samples/{0}/{1}/report_triage.json'.format(
                sample_id, task_id)
        )

    def sample_task_file(self, sample_id, task_id, filename):
        """
        Return a task file.

        Parameters:
            sample_id (str): The id of the sample
            task_id (str): The id of the task
            filename (str): The name of the file

        Returns:
            response (file):
                File object.
        """
        return self._req_file(
            "GET", "/v0/samples/{0}/{1}/{2}".format(
                sample_id, task_id, filename)
        )

    def sample_archive_tar(self, sample_id):
        """
        Return a tar achive of a sample.

        Parameters:
            sample_id (str): The id of the sample

        Returns:
            response (file):
                File object.
        """
        return self._req_file(
            "GET", "/v0/samples/{0}/archive".format(sample_id)
        )

    def sample_archive_zip(self, sample_id):
        """
        Return a zip achive of a sample.

        Parameters:
            sample_id (str): The id of the sample

        Returns:
            response (file):
                File object.
        """
        return self._req_file(
            "GET", "/v0/samples/{0}/archive.zip".format(sample_id)
        )

    def create_profile(self, name, tags, network, timeout):
        """
        Create a new profile.

        Parameters:
            name (str): The name of the profile
            tags (list): Tags for the profile, list of strings
            timeout (int): The timeout of the profile

        Returns:
            response (dict):
                {}, empty dict
        """
        return self._req_json("POST", "/v0/profiles", data={
            "name": name,
            "tags": tags,
            "network": network,
            "timeout": timeout
        })

    def delete_profile(self, profile_id):
        """
        Delete profile.

        Parameters:
            profile_id (str): The name or id of the profile

        Returns:
            response (dict):
                {}, empty dict
        """
        return self._req_json('DELETE', '/v0/profiles/{0}'.format(profile_id))

    def profiles(self, max=20):
        """
        Returns a Paginator object with profiles.

        Parameters:
            max (int): The maximum amount of profiles to return

        Returns:
            Paginator (object):
                Loop over this object to get the profiles
        """
        return Paginator(self, '/v0/profiles', max)

    def sample_events(self, sample_id):
        """
        Stream events of a running sample.

        Parameters:
            sample_id (str): The id of the sample

        Returns:
            yield of dict events
        """
        events = self._new_request("GET", "/v0/samples/"+sample_id+"/events")
        with Session() as s:
            res = s.send(events.prepare(), stream=True)
            for line in res.iter_lines():
                if line:
                    yield json.loads(line)

def PrivateClient(token):
    return Client(token, "https://private.tria.ge/api")

class ServerError(Exception):
    def __init__(self, err):
        b = err.response.json()
        self.status = err.response.status_code
        self.kind = b['error']
        self.message = b['message']

    def __str__(self):
        return 'triage: {0} {1}: {2}'.format(
            self.status, self.kind, self.message)


def encode_multipart_formdata(fields):
    boundary = binascii.hexlify(os.urandom(16)).decode('ascii')

    body = BytesIO()
    for field, value in fields.items(): # (name, file)
        if isinstance(value, tuple):
            filename, file = value
            body.write('--{boundary}\r\nContent-Disposition: form-data; '
                       'filename="{filename}"; name=\"{field}\"\r\n\r\n'
                .format(boundary=boundary, field=field, filename=filename)
                .encode('utf-8'))
            b = file.read()
            if isinstance(b, str):  # If the file was opened in text mode
                b = b.encode('ascii')
            body.write(b)
            body.write(b'\r\n')
        else:
            body.write('--{boundary}\r\nContent-Disposition: form-data;'
                       'name="{field}"\r\n\r\n{value}\r\n'
                .format(boundary=boundary, field=field, value=value)
                .encode('utf-8'))
    body.write('--{0}--\r\n'.format(boundary).encode('utf-8'))
    body.seek(0)

    return body, "multipart/form-data; boundary=" + boundary



# Copyright (C) 2020-2021 Hatching B.V.
# All rights reserved.

import sys
import os
import time
import json

import click
import appdirs

from triage import Client
from triage.client import ServerError
from cli.tui import prompt_select_options

def token_file():
    return os.path.join(appdirs.user_config_dir(), "triage.conf")

def client_from_env():
    tokenfile = token_file()
    if not os.path.exists(tokenfile):
        print("Please authenticate")
        sys.exit()
        return

    with open(token_file(), "r") as f:
        for line in f:
            line = line.strip()
            if len(line) == 0 or line.startswith("#"):
                continue

            url, token = line.split(" ")
            return Client(token, root_url=url)

    print("%s is not formatted correctly" % tokenfile)
    sys.exit()

@click.group()
def cli():
    pass

@cli.command()
@click.argument("token")
@click.option("-u", "--url", default="https://api.tria.ge", help="The endpoint of your triage instance")
def authenticate(token, url):
    tokenfile = token_file()
    if os.path.exists(tokenfile):
        print("Tokenfile already exists, currenty appending tokens is not "
            "supported, please edit/remove: ", tokenfile)
        return

    with open(token_file(), "w") as f:
        f.write("%s %s" % (url, token))

def prompt_select_files(static):
    print("Please select the files from the archive to analyize/")
    print("Leave blank to continue with the emphasized files and automatic "
        "profiles.")

    selection = prompt_select_options(static["files"], key="filename")
    if len(selection) == 0:
        return [{
                "name": x["filename"],
                "path": x.get("relpath")
            } for x in static["files"] if x["selected"]], True
    return [{
            "name": static["files"][i]["filename"],
            "path": static["files"][i].get("relpath")
        } for i in selection], False

def prompt_select_profiles_for_files(profiles, pick):
    f = None
    if len(pick) > 1:
        f = lambda x : len(x) > 0

    rt = []
    for i in pick:
        print("Please select the profiles to use for", i["name"])
        selection = prompt_select_options(
            profiles,
            key="name",
            f=f
        )
        for choice in selection:
            rt.append({
                "profile": profiles[choice]["id"],
                "pick": i["path"]
            })
    return rt

def prompt_select_profile(c, sample):
    for events in c.sample_events(sample):
        if events["status"] == "pending":
            print("waiting for static analysis to finish")
        elif events["status"] == "static_analysis":
            break
        elif events["status"] == "failed":
            print("the sample is in a failed state")
            return
        else:
            print("the sample does not need a profile to be selected")
            return
    static = c.static_report(sample)
    pick = []
    default_selection = False
    if static["sample"]["kind"] == "url":
        pick.append({
            "name": static["sample"]["target"],
            "path": static["sample"]["target"]
        })
    elif len(static["files"]) == 1:
        pick.append({
            "name": static["files"][0]["filename"],
            "path": static["files"][0].get("relpath"),
        })
    else:
        pick, default_selection = prompt_select_files(static)

    # Fetch profiles before determining whether we should use automatic
    #  profiles. If no profiles are available, fall back to automatic profiles.
    profiles = [x for x in c.profiles()]
    default_selection = (len(profiles) == 0)

    profile_selections = []
    if not default_selection:
        profile_selections = prompt_select_profiles_for_files(profiles, pick)
        default_selection = (len(profile_selections) == 0)

    if default_selection:
        print("Using default selection.")
        c.set_sample_profile_automatically(
            sample,
            pick=[i["path"] for i in pick]
        )
        return

    c.set_sample_profile(sample, profile_selections)

@cli.command("submit")
@click.argument("target")
@click.option("-i", "--interactive", is_flag=True, help="Perform interactive"
" submission where you can manually select the profile and files")
@click.option("-p", "--profile", multiple=True, help="The profile names or IDs"
" to use")
def submit(target, interactive, profile):
    f, url = None, None
    if os.path.exists(target):
        f = target
    else:
        url = target

    if interactive and profile:
        print("--interactive and --profile are mutually exclusive")
        return

    c = client_from_env()
    if f:
        name = os.path.basename(f)
        r = c.submit_sample_file(
            name, open(f, "rb"),
            interactive=interactive,
            profiles=[{
                "profile": x
            } for x in profile]
        )
    elif url:
        r = c.submit_sample_url(
            url, interactive=interactive,
            profiles=[{
                "profile": x
            } for x in profile]
        )
    else:
        print("Please specify -f file or -u url")
        return

    print("Sample submitted")
    print("  ID:       %s" % r["id"])
    print("  Status:   %s" % r["status"])
    if f:
        print("  Filename: %s" % r["filename"])
    else:
        print("  URL:      %s" % r["url"])

    if interactive:
        time.sleep(2)
        prompt_select_profile(c, r["id"])

@cli.command("select-profile")
@click.argument("sample")
def select_profile(sample):
    c = client_from_env()
    prompt_select_profile(c, sample)

def paginator_format(c, i):
    target = i["url"] if i.get("url") else i.get("filename", "-")
    if i["status"] == "reported":
        try:
            overview = c.overview_report(i["id"])
        except ServerError:
            return
        if len(overview["analysis"].get("family", [])) >= 1:
            print("%s\t%s, %s\t%s" % (
                overview["analysis"]["score"],
                i["id"],
                overview["analysis"]["family"],
                target
            ))
        else:
            print("%s\t%s\t%s" % (
                overview["analysis"]["score"],
                i["id"],
                target
            ))
    else:
        print(".\t%s\t%s" % (
            i["id"],
            target
        ))

@cli.command("list")
@click.option("-n", default=20, help="The maximum number of samples to return")
@click.option("-p", "--public", is_flag=True, help="List public samples")
def list_samples(public, n):
    c = client_from_env()
    for i in c.public_samples(max=n) if public else c.owned_samples(max=n):
        paginator_format(c, i)

@cli.command("file")
@click.argument("sample")
@click.argument("task")
@click.argument("file")
@click.option("-o", "--output", help= "The path to where the "
    "downloaded file should be saved. If `-`, the file is copied to stdout")
def get_file(sample, task, file, output):
    c = client_from_env()
    f = c.sample_task_file(sample, task, file)
    if output == "-":
        print(f)
    if not output:
        output = "".join(x for x in file if x not in "\/:*?<>|")
    with open(output, "wb") as wf:
        wf.write(f)

@cli.command("archive")
@click.argument("sample")
@click.option("-f", "--format", default="tar",
    help="The archive format. Either \"tar\" or \"zip\"")
@click.option("-o", "--output", help="The target file name. If `-`, the file "
"is copied to stdout. Defaults to the sample ID with appropriate extension")
def archive(sample, format, output):
    c = client_from_env()
    if format == "tar":
        r = c.sample_archive_tar(sample)
    elif format == "zip":
        r = c.sample_archive_zip(sample)
    else:
        print("Use --format zip or tar")
        return

    if output == "-":
        print(r)
    elif output:
        with open(output, "wb") as wf:
            wf.write(r)
    else:
        with open("%s.%s" % (sample, format), "wb") as wf:
            wf.write(r)


@cli.command("delete")
@click.argument("sample")
def delete(sample):
    c = client_from_env()
    c.delete_sample(sample)

@cli.command("onemon.json")
@click.argument("sample")
@click.argument("tasks", nargs=-1)
def onemon(sample, tasks):
    c = client_from_env()
    for k in c.overview_report(sample)["tasks"]:
        if k["kind"] == "behavioral":
            if tasks and k["name"] not in tasks:
                continue
            for line in c.kernel_report(sample, k["name"]):
                print(json.dumps(line, separators=(',', ':')))

@cli.command("search", help="Use https://tria.ge/docs/cloud-api/samples/#get-search for query formats")
@click.argument("query")
@click.option("-n", default=20, help="The maximum number of samples to return")
def search(query, n):
    c = client_from_env()
    for i in c.search(query, n):
        paginator_format(c, i)

@cli.command("report")
@click.argument("sample")
@click.option("--static", is_flag=True, help="Query the static report")
@click.option("-t", "--task", help= "The ID of the report")
def report(sample, static, task):
    c = client_from_env()
    if static:
        print("~Static Report~")
        r = c.static_report(sample)
        for f in r["files"]:
            print("%s %s" % (
                f["filename"],
                "(selected)" if f["selected"] else "")
            )
            print("  md5:", f["md5"])
            print("  tags:", f.get("tags", []))
            print("  kind:", f["kind"])
    elif task:
        print("~%s Report~" % task)
        r = c.task_report(sample, task)
        err = r.get("errors")
        if err:
            print(err)
            return
        print(r["task"]["target"])
        print("  md5:", r["task"]["md5"])
        print("  score:", r["analysis"]["score"])
        print("  tags:", r["analysis"].get("tags", []))
    else:
        print("~Overview~")
        r = c.overview_report(sample)
        if r.get("errors"):
            print("Triage produced the following errors", r["errors"])
        print(r["sample"]["target"])
        print("  md5:", r["sample"]["md5"])
        print("  score:", r["analysis"]["score"])
        print("  family:", r["analysis"].get("family"))
        print("  tags:", r["analysis"].get("tags", []))
        print()
        for task in r.get("tasks", []):
            print(" ", task["name"])
            print("    score:", task.get("score", "N/A"))
            if task["kind"] != "static":
                print("    platform:", task["platform"])
            print("    tags:", task.get("tags", []))

@cli.command("create-profile")
@click.option("--name", required=True, help="The name of the new profile")
@click.option("--tags", required=True, help="A comma separated set of tags")
@click.option("--network", help="The network type to use. Either \"internet\","
    " \"drop\" or unset")
@click.option("--timeout", required=True, type=int,
    help="The timeout of the profile")
def create_profile(name, tags, network, timeout):
    c = client_from_env()
    r = c.create_profile(name, tags.split(","), network, timeout)
    print(r)

@cli.command("delete-profile")
@click.option("-p", "--profile", required=True,
    help="The name or ID of the profile")
def delete_profile(profile):
    c = client_from_env()
    r = c.delete_profile(profile)
    print(r)

@cli.command("list-profiles")
def list_profiles():
    c = client_from_env()
    for i in c.profiles():
        print(i["name"])
        print("  timeout:", i["timeout"])
        print("  network:", i["network"])
        print("  tags:", i.get("tags", []))
        print("  id:", i["id"])

if __name__ == "__main__":
    cli()



# Copyright (C) 2020 Hatching B.V.
# All rights reserved.

import json

def prompt_select_options(options, key, f=None):
    print("\nMake your selection by entering the numbers as listed below "
    "separated by spaces and finish with enter.\n")

    i = 0
    for option in options:
        em = " "
        if option.get("selected"):
            em = ">"
        print(em, i, option[key])
        i += 1

    selection = []
    choices = input("> ")
    for choice in choices.split(" "):
        if choice.strip() == "":
            continue

        try:
            choice = int(choice)
        except ValueError:
            print("Bad input ", choice)
            continue

        if choice < 0 or len(options) <= choice:
            print("Out of range ", choice)
            continue

        if choice in selection:
            continue

        selection.append(choice)

    if f:
        if not f(selection):
            return prompt_select_options(options, key,  f)

    if len(selection) > 0:
        print("You selected:")
    for choice in selection:
        print(" ", options[choice][key])
    return selection



# Copyright (C) 2020-2021 Hatching B.V.
# All rights reserved.

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="hatching-triage",
    version="0.1.4",
    author="Hatching B.V.",
    author_email="info@hatching.io",
    description="API client and CLI for Hatching Triage",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://hatching.io/",
    packages=setuptools.find_packages(),
    entry_points={
        "console_scripts": [
            "triage = cli.triage:cli",
        ],
    },
    install_requires=[
        "appdirs==1.4.4",
        "click==7.1.2",
        "requests==2.25.1"
    ],
    python_requires='>=3.6',
)



# Copyright (C) 2020-2021 Hatching B.V
# All rights reserved.

class Paginator:
    def __init__(self, client, path, max):
        self._client = client
        self._path = path
        self._offset = None
        self._limit = 200
        self._current_page = []
        self._eof = False
        self._max = max
        self._counter = 0

    def __iter__(self):
        return self

    def _fetch_next_page(self):
        if '?' in self._path:
            path = self._path + '&'
        else:
            path = self._path + '?'

        if self._max < self._limit:
            self._limit = self._max

        path = path + 'limit={0}'.format(self._limit)
        if self._offset is not None:
            path = path + '&offset={0}'.format(self._offset)

        resp = self._client._req_json('GET', path)
        self._current_page = resp['data']

        if resp.get('next'):
            self._offset = resp['next']
        else:
            self._offset = None
            self._eof = True

        return len(resp['data']) > 0

    def __next__(self):
        if self._counter == self._max:
            raise StopIteration

        if len(self._current_page) == 0:
            if self._eof:
                raise StopIteration
            if not self._fetch_next_page():
                raise StopIteration

        self._counter += 1
        return self._current_page.pop(0)
